"""Meta tools for integrations."""

import inspect


class MetaIntegration(type):
    """
    Metaclass for context-bound integrations.

    The purpose of this metaclass is to simplify writing integrations that
    need to attach to the application or request context.

    A typical pattern  when using an integration's method is to first check for
    an instance of the integration on the current context, and then call the
    method on that instance. In practice, this means either performing the
    context check  every time the method is used (yuck) or implementing
    module-level functions that wrap the integration class instance methods,
    and check the context  when called (also yuck).

    Classes that use this metaclass are expected to implement a classmethod
    called ``current_session()`` that returns an instance of the class. That
    method is responsible for getting the correct instance for the context.

    The net effect is that instance methods will be exposed as if they were
    class methods; behind the scenes, this metaclass will call
    ``current_session()`` and obtain the bound method from the returned
    instance.
    """

    def __getattribute__(self, key):
        """Get the attribute from the instance bound to the current context."""
        obj = super(MetaIntegration, self).__getattribute__(key)
        if inspect.isfunction(obj):
            return getattr(self.current_session(), key)
        return obj



"""Provides a base class for HTTP service integrations."""

from typing import Optional, Tuple, MutableMapping, List
import inspect
import requests
from urllib3.util.retry import Retry
from urllib.parse import urlparse, urlunparse, urlencode
import json

from flask import g, Flask, current_app

from . import status
from ..meta import MetaIntegration
import logging

from .exceptions import RequestFailed, RequestUnauthorized, RequestForbidden, \
    NotFound, BadRequest, SecurityException, ConnectionFailed, BadResponse

logger = logging.getLogger(__name__)

DEFAULT_ENDPOINT = 'http://fooservice:8000/'
DEFAULT_VERIFY = True


class HTTPIntegration(metaclass=MetaIntegration):
    """
    Integrate with a service over HTTP.

    Provides binding to the application context, so that connections can be
    re-used.
    """

    class Meta:
        """
        Additional info about this class.

        This should be overridden by any child class.
        """

        service_name = "base"

    def __init__(self, endpoint: str, verify: bool = True,
                 headers: dict = {}) -> None:
        """
        Initialize an HTTP session.

        Parameters
        ----------
        endpoint : str
            Base service URI.
        verify : bool
            Whether or not SSL certificate verification should enforced.
        headers : dict
            Headers to be included on all requests.

        """
        self._session = requests.Session()
        self._verify = verify
        self._retry = Retry(  # type: ignore
            total=10,
            read=10,
            connect=10,
            status=10,
            backoff_factor=0.5
        )
        self._adapter = requests.adapters.HTTPAdapter(max_retries=self._retry)
        self._session.mount(f'{urlparse(endpoint).scheme}://', self._adapter)
        if not endpoint.endswith('/'):
            endpoint += '/'
        self._endpoint = endpoint
        self._session.headers.update(headers)

    def _path(self, path: str, query: dict = {}) -> str:
        """Generate a full path for a request."""
        o = urlparse(self._endpoint)
        path = path.lstrip('/')
        return urlunparse((  # type: ignore
            o.scheme, o.netloc, f"{o.path}{path}",
            None, urlencode(query), None
        ))

    def _check_status(self, resp: requests.Response,
                      expected_code: List[int] = [status.HTTP_200_OK]) -> None:
        """Check for unexpected or errant status codes."""
        if resp.status_code >= status.HTTP_500_INTERNAL_SERVER_ERROR:
            raise RequestFailed(f'Status: {resp.status_code}', resp)
        elif resp.status_code == status.HTTP_401_UNAUTHORIZED:
            raise RequestUnauthorized(f'Not authorized', resp)
        elif resp.status_code == status.HTTP_403_FORBIDDEN:
            raise RequestForbidden(f'Forbidden', resp)
        elif resp.status_code == status.HTTP_404_NOT_FOUND:
            raise NotFound(f'No such resource: {resp.url}', resp)
        elif resp.status_code >= status.HTTP_400_BAD_REQUEST:
            raise BadRequest(f'Bad request: {resp.content}', resp)
        elif resp.status_code not in expected_code:
            raise RequestFailed(f'Unexpected code: {resp.status_code}', resp)

    def _parse_location(self, location: str) -> str:
        endpoint_path = urlparse(self._endpoint).path
        if self._endpoint in location:
            _, location = location.split(self._endpoint, 1)
        elif endpoint_path in location:
            _, location = location.split(endpoint_path, 1)
        return location

    def request(self, method: str, path: str, token: Optional[str] = None,
                expected_code: List[int] = [status.HTTP_200_OK],
                allow_2xx_redirects: bool = True,
                **kwargs) -> requests.Response:
        """Make an HTTP request with error/code handling."""
        if token is not None:
            if 'headers' not in kwargs:
                kwargs['headers'] = {}
            kwargs['headers'].update({'Authorization': token})
        try:
            resp = getattr(self._session, method)(self._path(path), **kwargs)
            logger.debug('Got response %s', resp)
        except requests.exceptions.SSLError as e:
            raise SecurityException('SSL failed: %s' % e) from e
        except requests.exceptions.ConnectionError as e:
            raise ConnectionFailed('Could not connect: %s' % e) from e

        # 200-series redirects are a common pattern on our projects, so these
        # should be supported.
        is_2xx = status.HTTP_200_OK < resp.status_code \
            and resp.status_code < status.HTTP_300_MULTIPLE_CHOICES
        if allow_2xx_redirects and is_2xx and 'Location' in resp.headers:
            loc = self._parse_location(resp.headers['Location'])
            logger.debug('Following 2xx redirect to %s', loc)
            resp = self.request('get', loc, token, expected_code,
                                allow_2xx_redirects)

        self._check_status(resp, expected_code)
        return resp

    def json(self, method: str, path: str, token: Optional[str] = None,
             expected_code: List[int] = [status.HTTP_200_OK], **kwargs) \
            -> Tuple[dict, int, MutableMapping]:
        """
        Perform an HTTP request to a JSON endpoint, and handle any exceptions.

        Returns
        -------
        dict
            Response content.
        int
            HTTP status code.
        dict
            Response headers.

        """
        response = self.request(method, path, token, expected_code, **kwargs)
        try:
            return response.json(), response.status_code, response.headers
        except json.decoder.JSONDecodeError as e:
            raise BadResponse(f'Could not decode', response) from e

    def get_status(self) -> dict:
        """Get the status of the file management service."""
        logger.debug('Get service status')
        content = self.json('get', 'status')[0]
        logger.debug('Got status response: %s', content)
        return content

    @classmethod
    def init_app(cls, app: Flask) -> None:
        """Set default configuration parameters for an application instance."""
        if not hasattr(cls, 'Meta'):
            raise NotImplementedError('Child class must have Meta class')
        name = cls.Meta.service_name.upper()
        app.config.setdefault(f'{name}_ENDPOINT', DEFAULT_ENDPOINT)
        app.config.setdefault(f'{name}_VERIFY', DEFAULT_VERIFY)

    @classmethod
    def get_session(cls, app: Optional[Flask] = None) -> 'HTTPIntegration':
        """Get a new session with the service."""
        if not hasattr(cls, 'Meta'):
            raise NotImplementedError('Child class must have Meta class')
        if app is None:
            app = current_app
        name = cls.Meta.service_name.upper()
        try:
            endpoint = app.config[f'{name}_ENDPOINT']
            verify = app.config[f'{name}_VERIFY']
        except KeyError as e:
            raise RuntimeError('Must call init_app() on app before use') from e
        logger.debug('Create %s session at endpoint %s', name, endpoint)
        return cls(endpoint, verify=verify)

    @classmethod
    def current_session(cls) -> 'HTTPIntegration':
        """Get or create an HTTP session for this context."""
        name = cls.Meta.service_name
        if not g:
            return cls.get_session()
        elif name not in g:
            setattr(g, name, cls.get_session())  # type: ignore
        return getattr(g, name)  # type: ignore



"""Template filters."""

import re
from functools import partial
from typing import Union
from flask import Flask
from jinja2 import Markup, escape
from arxiv.base.urls import urlize, url_for_doi, canonical_url, \
    clickthrough_url
from arxiv.util.tex2utf import tex2utf
from arxiv.taxonomy import get_category_display, get_archive_display, \
    get_group_display

JinjaFilterInput = Union[Markup, str]
"""
   Jinja filters will receive their text input as either
   a Markup object or a str. It is critical for proper escaping to
   to ensure that str is correctly HTML escaped.

   Markup is decoded from str so this type is redundant but
   the hope is to make it clear what is going on to arXiv developers.
"""


def abstract_lf_to_br(text: JinjaFilterInput) -> Markup:
    """Lines that start with two spaces should be broken."""
    if isinstance(text, Markup):
        etxt = text
    else:
        etxt = Markup(escape(text))

    # if line starts with spaces, replace the white space with <br />
    br = re.sub(r'((?<!^)\n +)', '\n<br />', etxt)
    dedup = re.sub(r'\n\n', '\n', br)  # skip if blank
    return Markup(dedup)


def f_tex2utf(text: JinjaFilterInput,
              letters: bool = True) -> Markup:
    """Return output of tex2utf function as escaped Markup."""
    if isinstance(text, Markup):
        return escape(tex2utf(text.unescape(), letters=letters))
    else:
        return Markup(escape(tex2utf(text, letters=letters)))


def register_filters(app: Flask) -> None:
    """
    Register base template filters on a Flask app.

    Parameters
    ----------
    app : :class:`Flask`

    """
    app.template_filter('abstract_lf_to_br')(abstract_lf_to_br)
    app.template_filter('urlize')(urlize)
    app.template_filter('tex2utf')(partial(f_tex2utf, letters=True))
    app.template_filter('tex2utf_no_symbols')(partial(f_tex2utf, letters=False))
    app.template_filter('canonical_url')(canonical_url)
    app.template_filter('url_for_doi')(url_for_doi)
    app.template_filter('clickthrough_url')(clickthrough_url)
    app.template_filter('get_category_display')(get_category_display)
    app.template_filter('get_archive_display')(get_archive_display)
    app.template_filter('get_group_display')(get_group_display)
